from edsl.utilities.interface import (
    print_dict_as_html_table,
    print_dict_with_rich,
    print_list_of_dicts_as_html_table,
    print_list_of_dicts_with_rich,
    print_table_with_rich,
    print_public_methods_with_doc,
)

from edsl.utilities.utilities import (
    create_valid_var_name,
    dict_to_html,
    hash_value,
    HTMLSnippet,
    is_notebook,
    is_gzipped,
    is_valid_variable_name,
    random_string,
    repair_json,
    shorten_string,
)
